
/*****************************************************************************
	hc2_count with tilted time HWINDOWS
	2nd Algorithm - no intermediate HWINDOWS
*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#define MAXINT INT_MAX

#include "hc2.h"

extern long N;		// current transaction number
extern long BSIZE;	// batch size

/***************************************************************************** 
hc2_type *hc2_init(double rho, double epsilon, long M)
INPUTS:
	rho - hCount parameter
	epsilon - hCount parameter
	M - largest item in the stream

OUTPUT:
	Pointer to initialized hc2_type struct

Creates an instance of the hc2_type struct and initializes everything
*****************************************************************************/
hc2_type *hc2_init(double rho, double epsilon, long M)
{
    double e;			// min x/ln(x) and x > 1  --> evaluates to the base of the natural logarithm
    int R_RANGE;		// range for random numbers a_i and b_i
   
        int i;
    S2_TYPE *aptr;

    hc2_type *result = calloc(1, sizeof(hc2_type));
    if (result == NULL) {
		printf("\ncalloc error");
		exit(100);
    }

    result->N = 0;			// total transactions
    result->M = M; // long M = 0xffff; in mainwin. è il MAX value nello stream.
    
	e = 2.718281828;
    //result->P = 1048583;	// big prime number
	//result->P = 1021;	
	result->P = 65539;
	//result->P = 16411;

    R_RANGE = result->M;
    
	result->m =  e / epsilon;		// length of hash table 
    result->h = (int) (log(-result->M / log(rho)));		// width of hash table (numero di hash functions)
    
	// Allocate memory for the counters S
	aptr = (S2_TYPE *) malloc(result->m * result->h * sizeof(S2_TYPE));
    if (aptr == NULL) {
		printf("\nmalloc() error.");
		exit(100);
    }

    // e / 0.01 * log(-0xffff / log(0.99)) * N_OF_BUCKETS
    
	result->S = (S2_TYPE **) malloc(result->m *sizeof(S2_TYPE));
    if (result->S == NULL) {
		printf("\nmalloc() error.");
		exit(200);
    }
    for (i = 0; i < result->m; i++) {
		result->S[i] = aptr + i * result->h;
    }
    
	// Set all counters to 0
	memset((void *) aptr, 0, result->m * result->h * sizeof(S2_TYPE) / sizeof(char));
    
        // for hcount*:

	// generate h pairs of random numbers a_i and b_i
	result->a = (int *) malloc(result->h * sizeof(int));
    if (result->a == NULL) {
		printf("\nmalloc() error.");
		exit(300);
    }
    
	result->b = (int *) malloc(result->h * sizeof(int));
    if (result->b == NULL) {
		printf("\nmalloc() error.");
		exit(400);
    }
    for (i = 0; i < result->h; i++) {
		result->a[i] = random(0, R_RANGE);
		result->b[i] = random(0, R_RANGE);
    }

	return (result);
}

/*****************************************************************************
void hc2_update(hc2_type *hc, int val, int ttype)
INPUTS:
	hc - pointer to the hCount struct
	k - the item in the stream
	ttype - transaction type: 1 for INSERT, -1 for DELETE

OUTPUT:
	NIL

Updates the hCount table with new transaction. 
*****************************************************************************/
void hc2_update(hc2_type *hc, int val, int ttype)
{
        int pos, i, j;
	int remainder;

	N++;
	/* if (ttype == INSERT) {
		N++;
	} else {
		N--;
	} */

	/* shift once every BSIZE transactions */
	if (N % BSIZE == 0) {
		for (i=0; i < hc->m; i++) {
			for (j=0; j < hc->h; j++) {
                                shiftIncrement(hc->S[i][j].c);
				remainder = N % (2 * BSIZE);
				if (remainder < BSIZE) {
					hc->S[i][j].c[0] = 0;
				} else {
					hc->S[i][j].c[1] = 0;
				}
			}
		}
  	}


	/* Update the 0 level counters */
	for (j=0; j < hc->h; j++) {
		pos = ((((long long) hc->a[j] * val + hc->b[j]) % hc->P) % hc->m);
		if (N % (2 * BSIZE) < BSIZE)
			hc->S[pos][j].c[0]++;
		else
			hc->S[pos][j].c[1]++;
		
		/* Insert Deletion code here */
	
	}
}

void shiftIncrement(int *c)
{
    int i;
	int powerValue, threshold;

    for (powerValue = HWINDOWS-2; powerValue > 0; powerValue--) {
                threshold = (int)pow(2.0, (double)powerValue) * BSIZE;
		//System.out.println("threshold: " + threshold + "current:" + current);
		if ((N > 0) && (N % threshold == 0)) {
			c[powerValue + 1] = 0;
			for (i = 0; i < powerValue + 1; i++)
				c[powerValue + 1] += c[i];
			break;
		}
    }
}

/*****************************************************************************
long hc2_query(hc2_type *hc, long start, long end, long val)
INPUTS:
	hc - pointer to the hCount struct
	start - tqmin
	end - tqmax
	val - item

OUTPUT:
	Count of "item" 

Counts the number of occurences of "item" in window (tqmin, tqmax)
*****************************************************************************/
long hc2_query(hc2_type *hc, long start, long end, long val)
{
        long min, pos, j;
	int powerValue;
	long finalCount, counterSize;
	CounterInfo info;

	min = MAXINT;
	for (j=0; j < hc->h; j++) {
		pos = ((((long long) hc->a[j] * val + hc->b[j]) % hc->P) % hc->m);

	    info.startIndex = start;
		info.endIndex = end;
		info.currentSum = 0;
		info.started = 0;
		info.finished = 0;

		finalCount = 0;
		counterSize = 0;
		powerValue = 0;
		//printf("data: %ld %ld %ld\n", c->numOfCounters, c->current, c->blockSize);
		for (powerValue = HWINDOWS-2; powerValue > 0; powerValue--) {
			counterSize = (int) pow(2.0, (double) powerValue) * BSIZE;
			//System.out.println("threshold: " + counterSize + "current:" + counter + "mod " + counter%(2*counterSize));
			if (N % (2 * counterSize) >= counterSize) {
				finalCount += calculateSum(powerValue + 1, counterSize, hc->S[pos][j].c, &info);
			}
		}
		//printf("finalCount1: %ld\n", finalCount);
		//for 1a
		counterSize = BSIZE;
		finalCount += calculateSum(0, counterSize, hc->S[pos][j].c, &info);

		// printf("finalCount2: %ld\n", finalCount);
		//for 1b
		if (N % (2 * counterSize) >= counterSize) {
			finalCount += calculateSum(1, counterSize, hc->S[pos][j].c, &info);
		//printf("finalCount3: %ld\n", finalCount);
		}
		
		if (finalCount < min) {
			min = finalCount;
		}
	}

	return(min);
}

long calculateSum(int counterIndex, long counterSize, int *c, CounterInfo *info)
{
    double startFraction, finalFraction;

	if (info->finished)
		return 0;

    info->currentSum += counterSize;
    //printf("info: %ld %ld %ld %d\n", info->currentSum, info->startIndex, counterSize, info->started);
    //starting the count, getting start fraction and possibly end fraction
    if (!info->started && (info->currentSum > info->startIndex)) {
		info->started = TRUE;

		startFraction =	(double) (info->currentSum - info->startIndex) / (double) counterSize;
		//printf("data: %ld %ld %ld %f\n", info->currentSum, info->startIndex, counterSize, startFraction);
		if (info->started && (info->currentSum > info->endIndex)) {
			info->finished = TRUE;

			if (info->endIndex >= N)
				return c[counterIndex];

			finalFraction = (double) (info->endIndex - (info->currentSum -  counterSize)) / (double) counterSize;
			return (long)((startFraction - (1 - finalFraction)) *  c[counterIndex]);
		}
		return (long)(startFraction * c[counterIndex]);
    }
    //getting end fraction for blocks starting from beginning of block
    if (info->started && (info->currentSum > info->endIndex)) {
		info->finished = TRUE;

		if (info->endIndex >= N)
			return c[counterIndex];

		finalFraction = (double) (info->endIndex - (info->currentSum - counterSize)) / (double) counterSize;
		return (long)(finalFraction * c[counterIndex]);
    }
    //getting full blocks for blocks in the middle
    if (info->started && (info->currentSum <= info->endIndex)) {
		return c[counterIndex];
    }
    return 0;
}

/*****************************************************************************
int hc2_size(hc2_type * hc)
INPUTS:
	hc - pointer to the hCount structure

OUTPUT:
	Space occupied by the hCount structure in bytes
*****************************************************************************/
int hc2_size(hc2_type * hc)
{
    int size;
    size = hc->m * hc->h * sizeof(S2_TYPE);
    return (size);
}


/*****************************************************************************
void hc2_print(hc2_type *hc, int id)
INPUTS:
	hc - pointer to the hCount structure
	id - ID of Tilted Time Window

OUTPUT:
	NIL

Prints relevant debug information about the hCount structure
*****************************************************************************/
void hc2_print(hc2_type *hc, int id)
{
    int i, j;

    for (i = 0; i < hc->m; i++) {
		printf("\n");
		for (j = 0; j < hc->h; j++)
			printf("%d ", hc->S[i][j].c[id]);
    }

    for (i = 0; i < hc->h; i++) {
		printf("\n%d, a=%d, b=%d", i, hc->a[i], hc->b[i]);
    }

    printf("\nm = %d", hc->m);
    printf("\nh = %d", hc->h);
    printf("\nP = %d", hc->P);
    printf("\nwin_Print -- END");

}

/*****************************************************************************
void hc2_destroy(hc2_type * hc)
INPUTS:
	hc - pointer to the hCount structure

OUTPUT:
	NIL

Frees up memory used by the hCount structure
*****************************************************************************/
void hc2_destroy(hc2_type * hc)
{
    free(hc->a);
    free(hc->b);
    free(hc->S);
}

