
/*****************************************************************************
	TiTiCount with subtraction algorithm. The query part
	01-NOV-2007
*****************************************************************************/

#include <stdio.h>
#include <math.h>

#include <assert.h>
#include "histogram_sub.h"



#include <limits.h>
#define MAXINT INT_MAX

#define	FALSE	0
#define	TRUE	1

extern long N;		// current transaction count
extern long BSIZE;		// batch size

long histogram_num_transactions(int id);




histogram_struct *histogram_my_query(histogram_hc_type *hc, int id, long item)
{


    int j, pos;
    histogram_struct *s = NULL;
    int c;



    for (j = 0; j < hc->h; j++) {
        pos = ((((long long) hc->a[j] * item + hc->b[j]) % hc->P) % hc->m);

        if(!s) {
            s = &hc->S[pos][j].histogram[id];
            c = hc->S[pos][j].c[id];
            continue;
        }

        if(hc->S[pos][j].c[id] < c) {
            s = &hc->S[pos][j].histogram[id];
            c = hc->S[pos][j].c[id];
        }


    }

    return(s);
}

/*
histogram_struct *histogram_my_query(histogram_hc_type *hc, int id, long item)
{


    int min = INT_MAX;
    int j, pos, c;
    histogram_struct *s = NULL;

    for (j = 0; j < hc->h; j++) {
        pos = ((((long long) hc->a[j] * item + hc->b[j]) % hc->P) % hc->m);

        if(!s) {
            s = &hc->S[pos][j].histogram[id];
            continue;
        }

        if(hc->S[pos][j].histogram[id].f1 < s->f1 &&
           hc->S[pos][j].histogram[id].f2 < s->f2 &&
           hc->S[pos][j].histogram[id].f3 < s->f3 &&
           hc->S[pos][j].histogram[id].f4 < s->f4) {
            s = &hc->S[pos][j].histogram[id];
        }


    }

    return(s);
} */

double histogram_overlaps(long tqmin, long tqmax, long hmin,long hmax) {
    int max_min, min_max;

    if(tqmin > hmin)
        max_min = tqmin;
    else
        max_min = hmin;


    if(tqmax < hmax)
        min_max = tqmax;
    else
        min_max = hmax;

    if(max_min < min_max) {
        double v = ((double)(min_max - max_min)) / ((double)(hmax - hmin));
        //  printf("\ntqmin=%d tqmax=%d hmin=%d hmax=%d\n", tqmin, tqmax, hmin, hmax);
        //  printf("max_min=%d min_max=%d [%d,%d] <=> [%d,%d] overlapping: %f\n",max_min, min_max, tqmin, tqmax, hmin, hmax, v);
        return (v);
    }
    else
        return 0;
}

double histogram4_freq(histogram_hc_type *hc, long tqmin, long tqmax, int ws, int item) {

    double frac =((double) (tqmax - tqmin) )/((double) histogram_num_transactions(ws));
    double hfrac = 0;

    // ecco come viene chiamata questa funzione:
    //    cnt =  histogram_freq(hc, tqmin, tqmax, ws, ITEM) * histogram_count_single_window(hc, ws, ITEM);
    // histogram_count_single_window conta la frequenza nella finestra completa, poi
    // histogram_freq determina quale frazione di questa frequenza considerare.
    // ossia, quando la query overlappa con la finestra, che la contiene completamente.

    // return frac;

    histogram_struct *s = histogram_my_query(hc, ws, item);

    if(!s || (s->f1 == s->f2 == s->f3 == s->f4)) {
        return frac; // assume uniform distrib.
    }

    assert(s->f1 >= 0 && s->f1 <=3 &&
           s->f2 >= 0 && s->f2 <=3 &&
           s->f3 >= 0 && s->f3 <=3 &&
           s->f4 >= 0 && s->f4 <=3);

    double sum = s->f1 + s->f2 + s->f3 + s->f4; // this is the pdf area

    // printf("sum: %d+%d+%d+%d\n", s->f1,s->f2,s->f3,s->f4);
    assert(sum >= 1);

    long first = histogram_min_boundary(ws);
    long last = histogram_min_boundary(ws) + histogram_num_transactions(ws);
    long width = last -first;


    double overlaps;
    double v = 0;


    if((overlaps = histogram_overlaps(tqmin,tqmax,first+0,first+width*1/4)) > 0) {
        v+= ((double)s->f4)  * overlaps;
    }

    if((overlaps = histogram_overlaps(tqmin,tqmax,first+width*1/4,first+width*2/4)) > 0) {
        v+=((double) s->f3) * overlaps;
    }

    if((overlaps = histogram_overlaps(tqmin,tqmax,first+width*2/4,first+width*3/4)) > 0) {
        v+=((double) s->f2) * overlaps;
    }

    if((overlaps = histogram_overlaps(tqmin,tqmax,first+width*3/4,first+width)) > 0) {
        v+=((double) s->f1) * overlaps;
    }

    hfrac = v/sum; // normalization of pdf region


    // printf("winningh=%d%d%d%d hfrac=%f frac=%f v=%f sum=%f\n", s->f1 , s->f2 , s->f3 ,s->f4,hfrac,frac,v,sum);
    // printf("using hfrac: hfrac=%f frac=%f v=%f sum=%d\n", hfrac,frac,v,sum);

    // printf("USING HISTOGRAM: counters=%d win=%d first=%d last=%d width=%d tqmin=%d tqmax=%d v=%f sum=%d hfrac=%f frac=%f\n",
    //      counters, ws, first, last, width, tqmin, tqmax, v, sum, hfrac, frac);

    return hfrac;

}


double histogram4_freq_overlapping(histogram_hc_type *hc, long tqmin, long tqmax, int ws, int item) {
    // overlapping window = ws-1


    double frac =((double) (tqmax - tqmin) )/ ((double)( histogram_num_transactions(ws) - histogram_num_transactions(ws-1)));
    double hfrac = 0;


    histogram_struct *s = histogram_my_query(hc, ws, item);

    if(!s || (s->f3 == s->f4)) {
        return frac; // assume uniform distrib.
    }

    assert(s->f1 >= 0 && s->f1 <=3 &&
           s->f2 >= 0 && s->f2 <=3 &&
           s->f3 >= 0 && s->f3 <=3 &&
           s->f4 >= 0 && s->f4 <=3);

    double sum = s->f3 + s->f4;
    //  printf("sum: %d+%d\n", s->f3,s->f4);

    assert(sum >= 1);

    long first = histogram_min_boundary(ws);
    long last = histogram_min_boundary(ws) + histogram_num_transactions(ws);
    long width = last -first;

    double v = 0;



    double overlaps;


    assert(histogram_overlaps(tqmin,tqmax,first+width*2/4,first+width*3/4) == 0);
    assert(histogram_overlaps(tqmin,tqmax,first+width*3/4,first+width) == 0);



    if((overlaps = histogram_overlaps(tqmin,tqmax,first+0,first+width*1/4)) > 0) {
        v+= ((double)s->f4)  * overlaps;
    }

    if((overlaps = histogram_overlaps(tqmin,tqmax,first+width*1/4,first+width*2/4)) > 0) {
        v+=((double) s->f3) * overlaps;
    }


    assert(v <= sum);
    hfrac = v/sum;


    //  printf("winningh=%d%d%d%d hfrac=%f frac=%f v=%f sum=%f\n", s->f1 , s->f2 , s->f3 ,s->f4,hfrac,frac,v,sum);

    // printf("USING HISTOGRAM: counters=%d win=%d first=%d last=%d width=%d tqmin=%d tqmax=%d v=%f sum=%d hfrac=%f frac=%f\n",
    //      counters, ws, first, last, width, tqmin, tqmax, v, sum, hfrac, frac);


    //    printf("v=%f sum=%f\n", v, sum);

    return hfrac;
}














double histogram2_freq(histogram_hc_type *hc, long tqmin, long tqmax, int ws, int item) {

    double frac =((double) (tqmax - tqmin) )/((double) histogram_num_transactions(ws));
    double hfrac = 0;

    // ecco come viene chiamata questa funzione:
    //    cnt =  histogram_freq(hc, tqmin, tqmax, ws, ITEM) * histogram_count_single_window(hc, ws, ITEM);
    // histogram_count_single_window conta la frequenza nella finestra completa, poi
    // histogram_freq determina quale frazione di questa frequenza considerare.
    // ossia, quando la query overlappa con la finestra, che la contiene completamente.

    // return frac;

    histogram_struct *s = histogram_my_query(hc, ws, item);

    if(!s || (s->f1 == s->f2)) {
        return frac; // assume uniform distrib.
    }


    double sum = s->f1 + s->f2; // this is the pdf area

    // printf("sum: %d+%d+%d+%d\n", s->f1,s->f2,s->f3,s->f4);
    assert(sum >= 1);

    long first = histogram_min_boundary(ws);
    long last = histogram_min_boundary(ws) + histogram_num_transactions(ws);
    long width = last -first;


    double overlaps;
    double v = 0;


    if((overlaps = histogram_overlaps(tqmin,tqmax,first+0,first+width*1/2)) > 0) {
        v+= ((double)s->f2)  * overlaps;
    }


    if((overlaps = histogram_overlaps(tqmin,tqmax,first+width*1/2,first+width)) > 0) {
        v+=((double) s->f1) * overlaps;
    }

    hfrac = v/sum; // normalization of pdf region


    // printf("winningh=%d%d%d%d hfrac=%f frac=%f v=%f sum=%f\n", s->f1 , s->f2 , s->f3 ,s->f4,hfrac,frac,v,sum);
    // printf("using hfrac: hfrac=%f frac=%f v=%f sum=%d\n", hfrac,frac,v,sum);

    // printf("USING HISTOGRAM: counters=%d win=%d first=%d last=%d width=%d tqmin=%d tqmax=%d v=%f sum=%d hfrac=%f frac=%f\n",
    //      counters, ws, first, last, width, tqmin, tqmax, v, sum, hfrac, frac);

    return hfrac;

}


double histogram2_freq_overlapping(histogram_hc_type *hc, long tqmin, long tqmax, int ws, int item) {
    // overlapping window = ws-1


    double frac =((double) (tqmax - tqmin) )/ ((double)( histogram_num_transactions(ws) - histogram_num_transactions(ws-1)));
    double hfrac = 0;

    return frac;
}















double histogram8_freq(histogram_hc_type *hc, long tqmin, long tqmax, int ws, int item) {

    double frac =((double) (tqmax - tqmin) )/((double) histogram_num_transactions(ws));
    double hfrac = 0;

    // ecco come viene chiamata questa funzione:
    //    cnt =  histogram_freq(hc, tqmin, tqmax, ws, ITEM) * histogram_count_single_window(hc, ws, ITEM);
    // histogram_count_single_window conta la frequenza nella finestra completa, poi
    // histogram_freq determina quale frazione di questa frequenza considerare.
    // ossia, quando la query overlappa con la finestra, che la contiene completamente.

    // return frac;

    histogram_struct *s = histogram_my_query(hc, ws, item);

    if(!s || (s->f1 == s->f2 == s->f3 == s->f4 == s->f5 == s->f6 == s->f7 == s->f8)) {
        return frac; // assume uniform distrib.
    }


    double sum = s->f1 + s->f2 + s->f3 + s->f4 + s->f5 + s->f6 + s->f7 + s->f8; // this is the pdf area

    // printf("sum: %d+%d+%d+%d\n", s->f1,s->f2,s->f3,s->f4);
    assert(sum >= 1);

    long first = histogram_min_boundary(ws);
    long last = histogram_min_boundary(ws) + histogram_num_transactions(ws);
    long width = last -first;


    double overlaps;
    double v = 0;



    if((overlaps = histogram_overlaps(tqmin,tqmax,first+0,first+width*1/8)) > 0) {
        v+= ((double)s->f8)  * overlaps;
    }

    if((overlaps = histogram_overlaps(tqmin,tqmax,first+width*1/8,first+width*2/8)) > 0) {
        v+=((double) s->f7) * overlaps;
    }

    if((overlaps = histogram_overlaps(tqmin,tqmax,first+width*2/8,first+width*3/8)) > 0) {
        v+=((double) s->f6) * overlaps;
    }

    if((overlaps = histogram_overlaps(tqmin,tqmax,first+width*3/8,first+width*4/8)) > 0) {
        v+=((double) s->f5) * overlaps;
    }

    if((overlaps = histogram_overlaps(tqmin,tqmax,first+width*4/8,first+width*5/8)) > 0) {
        v+=((double) s->f4) * overlaps;
    }

    if((overlaps = histogram_overlaps(tqmin,tqmax,first+width*5/8,first+width*6/8)) > 0) {
        v+=((double) s->f3) * overlaps;
    }

    if((overlaps = histogram_overlaps(tqmin,tqmax,first+width*6/8,first+width*7/8)) > 0) {
        v+=((double) s->f2) * overlaps;
    }

    if((overlaps = histogram_overlaps(tqmin,tqmax,first+width*7/8,first+width*8/8)) > 0) {
        v+=((double) s->f1) * overlaps;
    }

    hfrac = v/sum; // normalization of pdf region


    // printf("winningh=%d%d%d%d hfrac=%f frac=%f v=%f sum=%f\n", s->f1 , s->f2 , s->f3 ,s->f4,hfrac,frac,v,sum);
    // printf("using hfrac: hfrac=%f frac=%f v=%f sum=%d\n", hfrac,frac,v,sum);

    // printf("USING HISTOGRAM: counters=%d win=%d first=%d last=%d width=%d tqmin=%d tqmax=%d v=%f sum=%d hfrac=%f frac=%f\n",
    //      counters, ws, first, last, width, tqmin, tqmax, v, sum, hfrac, frac);

    return hfrac;

}


double histogram8_freq_overlapping(histogram_hc_type *hc, long tqmin, long tqmax, int ws, int item) {
    // overlapping window = ws-1


    double frac =((double) (tqmax - tqmin) )/ ((double)( histogram_num_transactions(ws) - histogram_num_transactions(ws-1)));
    double hfrac = 0;


    histogram_struct *s = histogram_my_query(hc, ws, item);

    if(!s || (s->f5 == s->f6 == s->f7 == s->f8)) {
        return frac; // assume uniform distrib.
    }

    double sum = s->f5 + s->f6 + s->f7 + s->f8;
    //  printf("sum: %d+%d\n", s->f3,s->f4);

    assert(sum >= 1);

    long first = histogram_min_boundary(ws);
    long last = histogram_min_boundary(ws) + histogram_num_transactions(ws);
    long width = last -first;

    double v = 0;



    double overlaps;


    if((overlaps = histogram_overlaps(tqmin,tqmax,first+0,first+width*1/8)) > 0) {
        v+= ((double)s->f8)  * overlaps;
    }

    if((overlaps = histogram_overlaps(tqmin,tqmax,first+width*1/8,first+width*2/8)) > 0) {
        v+=((double) s->f7) * overlaps;
    }

    if((overlaps = histogram_overlaps(tqmin,tqmax,first+width*2/8,first+width*3/8)) > 0) {
        v+=((double) s->f6) * overlaps;
    }
    if((overlaps = histogram_overlaps(tqmin,tqmax,first+width*3/8,first+width*4/8)) > 0) {
        v+=((double) s->f5) * overlaps;
    }


    assert(v <= sum);
    hfrac = v/sum;


    //  printf("winningh=%d%d%d%d hfrac=%f frac=%f v=%f sum=%f\n", s->f1 , s->f2 , s->f3 ,s->f4,hfrac,frac,v,sum);

    // printf("USING HISTOGRAM: counters=%d win=%d first=%d last=%d width=%d tqmin=%d tqmax=%d v=%f sum=%d hfrac=%f frac=%f\n",
    //      counters, ws, first, last, width, tqmin, tqmax, v, sum, hfrac, frac);


    //    printf("v=%f sum=%f\n", v, sum);

    return hfrac;
}


















/*****************************************************************************
long query(hc_type *hc, long tqmin, long tqmax, long ITEM)
INPUTS:
	hc - the hCount structure
	tqmin - lower transaction number
	tqmax - upper transaction number
	ITEM - the item to find

OUTPUT:
	Number of occurences of ITEM in the range (tqmin, tqmax) of transactions
*****************************************************************************/
long histogram_query(histogram_hc_type *hc, long tqmin, long tqmax, long ITEM)
{
    long cnt;
    int im_small;					// immediate smaller window (see Base Case)
    int ws;
    long twsmin;
    double frac, qdiff;

    qdiff = (tqmax-tqmin);

    ws = histogram_smallest_window(tqmax);	// find smallest window containing tqmax
    twsmin = histogram_min_boundary(ws);		// find starting transaction number in ws
    //	printf("\ntqmax = %ld, smallest_window = %d, min_boundary = %ld", tqmax, ws, twsmin);

    /****************** BASE CASE **************************

						 tqmin  tqmax
    OLD                    |      |                     NEW
	----------|----------|-{------}-|----------|----------|
                                                 |_____________________| <-- ws
						 |          
						 |          |__________| <-- immediate smaller overlapping window
					   twsmin		

	********************************************************/

    if (twsmin <= tqmin) {	// base case
        //printf("\nSUBQUERY: %ld, %ld", tqmin, tqmax);

        im_small = histogram_check_overlapping(ws);	// check if there is smaller overlapping window


        if (im_small != -1) {
            assert(ws -1 == im_small); //this always holds!
            //printf("overlapping: %d %d\n", ws, im_small);
            //  frac = qdiff / (histogram_num_transactions(ws) - histogram_num_transactions(im_small));
            cnt = histogram4_freq_overlapping(hc, tqmin, tqmax, ws, ITEM)*  histogram_count_with_subtraction(hc, ws, im_small, ITEM);
            //printf("\nfrac = %f", frac);
            //printf("-- SUBTRACTION NEEDED. \nwin%d - win%d. COUNT: %ld", ws, im_small, cnt);
        } else {
            // frac = qdiff / num_transactions(ws);
            // histogram_freq(hc, tqmin, tqmax, ws, ITEM) *
            cnt =  histogram4_freq(hc, tqmin, tqmax, ws, ITEM) * histogram_count_single_window(hc, ws, ITEM);
            //printf("\nfrac = %f", frac);
            //printf("-- NO SUBTRACTION. win%d. COUNT: %ld", ws, cnt);
        }
        return(cnt);
    } else {	// split into smaller windows recursively

        return(histogram_query(hc, twsmin, tqmax, ITEM) + histogram_query(hc, tqmin, twsmin-1, ITEM));

    }

}

long histogram_num_transactions(int id)
{
    if (id==0) {
        id = 1;
    }
    return((long)(BSIZE * pow(2.0, id-1)));
}

/*****************************************************************************
int smallest_window(long tran)
INPUTS:
	tran - transaction number

OUTPUT:
	The ID of the smallest window which contains the transaction tran
*****************************************************************************/
int histogram_smallest_window(long tran)
{
    /* This list will store the starting transaction number for each window.
		Assume 20 windows for the moment, tidy up later */
    long wfirst[20];		 // xxx will work only if there are 20 windows.
    int cnt=0;

    histogram_get_window_firsts(wfirst);

    while (wfirst[cnt] > tran) {
        cnt++;
    }

    return(cnt);
}


/*****************************************************************************
void get_window_firsts(long *wfirst)
INPUTS:
	wfirst - array to store the first transaction in each window
	
	uses N, the global current transaction count
	uses BSIZE, the global batch size

OUTPUT:
	NIL. 

Generates the wfirst array.
*****************************************************************************/
void histogram_get_window_firsts(long *wfirst)
{
    int k;
    int num;	// number of windows in use
    num = (int)((double)log((N-1)/BSIZE)/log(2.0)+1);		// log to the base 2
    //printf("\nNo. of windows: %d", num);

    for (k=0; k<=num; k++) {
        wfirst[k] = histogram_min_boundary(k);
        //printf("\nwfirst[%d] = %ld", k, wfirst[k]);
    }
    //getchar();
}


/*****************************************************************************
long min_boundary(int wID);
INPUTS:
	wID - window ID

	uses N, the global current transaction count
	uses BSIZE, the global batch size

OUTPUT:
	Smallest transaction number stored in window 'wID' (which is also the 
	min boundary of wID)
*****************************************************************************/
long histogram_min_boundary(int wID)
{
    int p, q;

    p = (N-1)/BSIZE;

    if (wID == 0) {		// current window - special case
        return(p * BSIZE + 1); // that, is N. wrong?
    } else {
        q = p % (long)pow(2, wID-1);
        return((p-q-pow(2,wID-1)) * BSIZE + 1);
    }
}


/*****************************************************************************
int check_overlapping(int wID)
INPUTS:
	wID - window ID

	uses BSIZE, the global batch size

OUTPUT:
	-1 if there is no smaller window overlapping with wID
	ELSE,
	ID of the immediate smaller window overlapping with wID
*****************************************************************************/
int histogram_check_overlapping(int wID)
{
    int FOUND = FALSE;
    int i;
    long last;	// last transaction in window wID

    last = histogram_min_boundary(wID) + pow(2, wID-1) * BSIZE - 1;

    for (i=wID-1; i>0; i--) {
        if (histogram_min_boundary(i) < last) {
            FOUND = TRUE;
            break;
        }
    }

    if (FOUND) {
        return(i);
    } else {
        return(-1);
    }
}


/*****************************************************************************
long count_single_window(hc_type *hc, int wID, long ITEM)
INPUTS:
	hc - the hCount structure
	wID - ID of window
	ITEM - item to count

OUTPUT:
	count of ITEM in wID
*****************************************************************************/
long histogram_count_single_window(histogram_hc_type *hc, int wID, long ITEM)
{
    return(histogram_hc_query(hc, wID, ITEM));
}


/*****************************************************************************
long count_with_subtraction(hc_type *hc, int wID1, int wID2, long ITEM)
INPUTS:
	hc - the hCount structure
	wID1 - ID of larger window
	wID2 - ID of smaller window (overlapping with wID1)
	ITEM - item to count

OUTPUT:
	count of ITEM in wID1 minus count of ITEM in wID2
*****************************************************************************/
long histogram_count_with_subtraction(histogram_hc_type *hc, int wID1, int wID2, long ITEM)
{
    return(histogram_count_single_window(hc, wID1, ITEM) - histogram_count_single_window(hc, wID2, ITEM));
}


/*****************************************************************************
void check_query(long tqmin, long tqmax)
INPUTS:
	tqmin - lower transaction number
	tqmax - upper transaction number

OUTPUT:
	NIL

Makes sure that the query is valid and that we can answer the query
based on the history stored so far. If not, print error message and quit.
*****************************************************************************/
void histogram_heck_query(long tqmin, long tqmax)
{

}
