#include "ssw2.h"

#include <stdio.h>
#include <stdlib.h>
//#include <malloc.h>
#include <math.h>
#include <string.h>

extern long N;		// current transaction number
extern long BSIZE;	// batch size


int smallest_window(long tran);
void get_window_firsts(long *wfirst);
long min_boundary(int wID);


SSW2_type * SSW2_Init(double epsilon)
{
    int i;

    SSW2_type *result = (SSW2_type *)calloc(1, sizeof(SSW2_type));

    result->epsilon = epsilon;

    for(i = 0; i < HWINDOWS; i++) {
        result->ss[i] = SS_Init(epsilon);
    }

    // printf("SS using %d counters.\n", result->ss[0]->numc);

    return result;
}



void SSW2_Update(SSW2_type *ssw2, int item, int ttype)
{
    int i,j;

    N++;
    // printf("at %d val = %d\n", N, item);

    SS_Update(ssw2->ss[0],item,ttype);


    int bn = N/BSIZE;

    if ((N -1) % BSIZE == 0) { // shift! //XXX non so perchè -1, cmq in titicount+ fa cosi'

        for(i = HWINDOWS-2;i>=0;i--) {

            if(bn % ((int)pow(2,i)) == 0) {

                //      printf("N = %d BN = %d shifting 0...%d to %d\n", N, bn, i, i+1);

                // reset counters in window i+1
                SS_Destroy(ssw2->ss[i+1]);
                ssw2->ss[i+1] = SS_Init(ssw2->epsilon);

                SS_bucket *b;
                SS_counter *c;

                //   int cnt = 0;

                for(j  = 0; j <= i; j++) {
                    for (b=ssw2->ss[j]->bfirst; b != NULL; b = b->next) {
                        if(b->value != 0)
                            for (c=b->ctr; c != NULL; c = c->next) {
                            // cnt += b->value;
                            SS_Add(ssw2->ss[i+1], c->id, c->bucket->value);
                        }
                    }
                }

                //      printf("inc to bucket %d cnt: %d\n", i+1, cnt);
                //      printf("bucket %d is empty: %d\n", i+i, ssw2->ss[i+1]->empty);

                //    break;
            }
        }


        SS_Destroy(ssw2->ss[0]); // xxx ho aggiunto queste due righe
        ssw2->ss[0] = SS_Init(ssw2->epsilon); //

    }

}


void update_freqs_sub(SSW2_type *ssw2, SS_type *result, double frac, int wid1, int wid2) {
    SS_bucket *b1,*b2;
    SS_counter *c1,*c2;
    int found;

    for (b1=ssw2->ss[wid1]->bfirst; b1 != NULL; b1 = b1->next) {
        if(b1->value != 0) {
            for (c1=b1->ctr; c1 != NULL; c1 = c1->next) {
                found = 0;

                for (b2=ssw2->ss[wid2]->bfirst; b2 != NULL; b2 = b2->next) {
                    if(b2->value != 0) {
                        for (c2=b2->ctr; c2 != NULL; c2 = c2->next) {
                            if(c1->id == c2->id) {
                                found = 1;
                                ssw2->subtractionFound++;
                                SS_Add(result, c1->id, (c1->bucket->value  - c2->bucket->value) * frac);
                            }
                        }
                    }
                }

                if(!found) {
                    ssw2->subtractionNotFound++;
                    SS_Add(result, c1->id, c1->bucket->value  * 0.5 * frac);
                }

            }
        }
    }

}

void update_freqs(SSW2_type *ssw2, SS_type *result, double frac, int wid) {
    SS_bucket *b;
    SS_counter *c;

    for (b=ssw2->ss[wid]->bfirst; b != NULL; b = b->next) {
        if(b->value != 0)
            for (c=b->ctr; c != NULL; c = c->next) {
            SS_Add(result, c->id, c->bucket->value * frac);
        }
    }


}

SS_type *SSW2_freqs(SSW2_type *ssw2, long tqmin, long tqmax) {
    SS_type *result = SS_Init(1.0/(double)0xffff);

    ssw2->subtractionFound = 0;
    ssw2->subtractionNotFound = 0;

    recursive_freqs(ssw2, result, tqmin, tqmax);

    printf("SSW2 subtractionFound(\%): %f\n", ssw2->subtractionFound / (ssw2->subtractionFound+ssw2->subtractionNotFound));

    return result;
}

int recursive_freqs(SSW2_type *ssw2, SS_type *result, long tqmin, long tqmax)
{
    SS_bucket *b;
    SS_counter *c;

    int im_small;					// immediate smaller window (see Base Case)
    int ws;
    long twsmin;
    double frac, qdiff;

    check_query(tqmin, tqmax);		// check validity of query
    qdiff = (tqmax-tqmin);

    ws = smallest_window(tqmax);	// find smallest window containing tqmax
    twsmin = min_boundary(ws);		// find starting transaction number in ws


    if(twsmin <= tqmin) {

        im_small = check_overlapping(ws);	// check if there is smaller overlapping window
        if (im_small != -1) {
            frac = qdiff / (num_transactions(ws) - num_transactions(im_small));
            update_freqs_sub(ssw2,result,frac,ws,im_small);
        } else {
            frac = qdiff / num_transactions(ws);
            update_freqs(ssw2,result,frac,ws);
        }
        return(0);

        // base case
    } else {
        return (recursive_freqs(ssw2,result,twsmin,tqmax) + recursive_freqs(ssw2,result,tqmin,twsmin-1));
    }

}


